(require :quickdraw)
(require :ff)

(defvar *help-symbols*)
(defvar *exported-symbols*)
(defvar *exported-but-no-doc*)
(defvar *exported-but-no-doc-ccl*)
(defvar *exported-but-no-doc-cl*)
(defvar *doc-but-not-exported*)
(defvar *doc-but-not-exported-ccl*)
(defvar *doc-but-not-exported-cl*)

(defun external-symbols (package)
  (let (res)
    (do-external-symbols (sym package) (push sym res))
    res))

(defun syms-in-package (syms package)
  (setq package (find-package package))
  (let (res)
    (dolist (sym syms)
      (when (and (eq package (symbol-package sym))
                 (not (memq (aref (string sym) 0) '(#\$ #\_))))
        (push sym res)))
    (nreverse res)))

(defun undocumented (&optional help-syms (string "doc strings"))
  (setq *help-symbols*
        (or help-syms
            (let (res)
              (maphash
               #'(lambda (sym ignore)
                   (declare (ignore ignore))
                   (push sym res))
               *fast-help*)
              res)))
  (setq *exported-symbols*
        (nconc
         (external-symbols :ccl)
         (external-symbols :cl)))
  (setq *exported-but-no-doc*
        (sort (set-difference *exported-symbols* *help-symbols*) #'string-lessp))
  (setq *doc-but-not-exported*
        (sort (set-difference *help-symbols* *exported-symbols*) #'string-lessp))
  (setq *exported-but-no-doc-ccl* (syms-in-package *exported-but-no-doc* :ccl))
  (setq *exported-but-no-doc-cl* (syms-in-package *exported-but-no-doc* :cl))
  (setq *doc-but-not-exported-ccl* (syms-in-package *doc-but-not-exported* :ccl))
  (setq *doc-but-not-exported-cl* (syms-in-package *doc-but-not-exported* :cl))
  (let ((w (make-instance 'fred-window)))
    (format w "These ~d CCL package symbols are exported but not in the ~a:~%"
            (length *exported-but-no-doc-ccl*) string)
    (dolist (sym *exported-but-no-doc-ccl*) (print sym w) (force-output w))
    (format w "~%~%These ~d CL package symbols are exported but not in the ~a:~%"
            (length *exported-but-no-doc-cl*) string)
    (dolist (sym *exported-but-no-doc-cl*) (print sym w) (force-output w))
    (format w "~%~%These ~d CCL package symbols are in the ~a but not exported:~%"
            (length *doc-but-not-exported-ccl*) string)
    (dolist (sym *doc-but-not-exported-ccl*) (print sym w) (force-output w))
    (format w "~%~%These ~d CL package symbols are in the ~a but not exported:~%"
            (length *doc-but-not-exported-cl*) string)
    (dolist (sym *doc-but-not-exported-cl*) (print sym w) (force-output w))
    w))

; Filters input-stream to output-stream.
; Writes only the first word of each line that starts with a
; CAPITALIZED word.
; Use to turn the pretty manual index into a list of symbols.
(defun get-manual-entries (input-stream output-stream)
  (file-position input-stream 0)
  (let ((string (make-array 100
                            :fill-pointer 0
                            :adjustable t
                            :element-type 'character))
        upper-case-p)
    (flet ((read-to-newline (reader reader-arg)
             (loop
               (when (or (stream-eofp input-stream)
                         (eql (the character (funcall reader reader-arg)) #\newline))
                 (return)))))
      (multiple-value-bind (reader reader-arg) (stream-reader input-stream)
        (loop
          (when (stream-eofp input-stream) (return))
          (setf (fill-pointer string) 0
                upper-case-p nil)
          (loop
            (let ((char (funcall reader reader-arg)))
              (declare (type character char))
              (when (or (eql char #\space) (eql char #\newline))
                (when upper-case-p
                  (write-string string output-stream)
                  (terpri output-stream)
                  (force-output output-stream))
                (unless (eql char #\newline) (read-to-newline reader reader-arg))
                (return))
              (when (lower-case-p char)
                (read-to-newline reader reader-arg)
                (return))
              (when (upper-case-p char)
                (setq upper-case-p t))
              (vector-push-extend char string))))))))


(defun check-keywords (strings)
  (let (res)
    (dolist (string strings)
      (unless (find-symbol string :keyword)
        (push string res)))
    (sort res 'string-lessp)))

; routines to compare real arglists to those in the
; help file or documentation
(defun check-arglists (&optional (help-file-arglist 'arglist))
  (require :quickdraw)
  (require :ff)
  (let (res)
    (do-external-symbols (sym :ccl)
      (let ((diff (check-symbol-arglist sym help-file-arglist)))
        (when diff (push diff res))))
    (do-external-symbols (sym :cl)
      (let ((diff (check-symbol-arglist sym help-file-arglist)))
        (when diff (push diff res))))
    (sort res #'string-lessp :key #'car)))

(defun check-symbol-arglist (sym &optional (help-file-arglist 'arglist))
  (unless (or (macro-function sym) (special-form-p sym))
    (multiple-value-bind (help help?) (funcall help-file-arglist sym)
      (multiple-value-bind (real real?) (arglist sym nil nil)
        (setq help (canonicalize-arglist help)
              real (canonicalize-arglist real))
        (flet ((bad ()
                 (return-from check-symbol-arglist
                   (list sym help real))))
          (declare (dynamic-extent bad))
          (when (and help? real?)
            (do* ((help-tail help (cdr help-tail))
                  (real-tail real (cdr real-tail))
                  (h (car help-tail))
                  (r (car real-tail)))
                 ((or (null help-tail) (null real-tail))
                  (unless (and (null help-tail) (null real-tail))
                    (bad)))
              (when (or (memq h lambda-list-keywords)
                        (memq r lambda-list-keywords))
                (unless (eq h r) (bad))
                (when (eq h '&key)
                  (unless (equal (cdr help-tail) (cdr real-tail))
                    (bad))
                  (return nil))))))))))

(defun canonicalize-arglist (arglist)
  (let ((key-tail (memq '&key arglist)))
    (when key-tail
      (setf (cdr key-tail) (sort (cdr key-tail) #'string-lessp))))
  arglist)

(defun get-doc-file-arglists (file)
  (let* ((w (ed file))
         (filename (file-namestring file))
         (buf (fred-buffer w))
         (b (make-mark buf 0))
         pos end-pos arg arglist result)
    (loop
        (setq pos (buffer-forward-search b "["))
        (unless pos (return))
        (set-mark b pos)
        (ignore-errors
         (setq end-pos (buffer-forward-search b "]
"))
         (when (and end-pos (< (buffer-line-start b end-pos -1) pos))
           (setq end-pos (buffer-bwd-skip-wsp b (- end-pos 2)))
           (let ((type (buffer-substring b pos end-pos)))
             (unless (member (read-from-string type) '(continued))
               (setq pos (buffer-line-start b pos))
               (let ((sym (buffer-read b pos)))
                 (setq pos (buffer-line-start b pos 1))
                 (setq arglist t)
                 (block get-arglist
                   (handler-bind ((error
                                   #'(lambda (condition)
                                       (declare (ignore condition))
                                       (set-mark b end-pos)
                                       (setq arglist t)
                                       (return-from get-arglist))))
                     (when (buffer-substring-p b "syntax" pos)
                       (incf pos 6)
                       (set-mark b pos)
                       (setq end-pos (buffer-line-start b pos 1))
                       (when end-pos
                         (multiple-value-bind (sym2 pos)
                                              (buffer-read b (buffer-position b) end-pos)
                           (unless (eq sym sym2)
                             (cerror "continue" "Second symbol disagrees: ~s ~s"
                                     sym sym2))
                           (setq arglist nil)
                           (loop
                             (set-mark b pos)
                             (multiple-value-setq (arg pos) (buffer-read b pos end-pos))
                             (unless pos (return))
                             (push arg arglist)))))))
                 (push (list sym (if (listp arglist) (nreverse arglist) t)
                             type filename) result)))))))
    (window-close w)
    (nreverse result)))

(defun get-all-doc-file-arglists (&optional (dir (choose-directory-dialog)))
  (let ((files (directory (merge-pathnames "*" dir)))
        res)
    (dolist (file files)
      (setq res (nconc (get-doc-file-arglists file) res)))
    (sort res 'string< :key 'car)))

(defun bad-doc-types (&aux res)
  (dolist (doc *doc-file-arglists*)
    (let ((sym (car doc))
          (type (third doc)))
      (unless
        (case (read-from-string type)
          ((primary generic)
           (typep (fboundp sym) 'generic-function))
          ((special variable)
           (if (equalp type "special form")
             (special-form-p sym)
             (eq (variable-information sym) :special)))
          (class
           (find-class sym nil))
          (condition
           (let ((class (find-class sym nil)))
             (and class
                  (memq (find-class 'condition) (class-precedence-list class)))))
          (function
           (let ((f (fboundp sym)))
             (and f (typep f 'function) (not (typep f 'generic-function)))))
          ((macro macro-definition)
           (macro-function sym))
          (otherwise t))
        (push doc res))))
  (nreverse res))

#|
(defparameter *bad-arglists*
  (check-arglists))

(let ((w (make-instance 'fred-window :window-title "Bad Arglists")))
  (pprint *bad-arglists* w)
  (fred-update w))

(defparameter *doc-file-arglists* (get-all-doc-file-arglists))

(defparameter *doc-file-definition-types*
  (sort (delete-duplicates (mapcar 'third *doc-file-arglists*) :test 'string-equal)
        'string-lessp))
  
(defparameter *unexported-doc-file-symbols*
  (let ((res nil))
    (require :quickdraw)
    (require :ff)
    (dolist (s *doc-file-arglists*)
      (let ((sym (car s)))
        (multiple-value-bind (sym2 type)
                             (find-symbol (symbol-name sym) (symbol-package sym))
          (declare (ignore sym2))
          (unless (eq type :external) (push s res)))))
    (nreverse res)))

(defparameter *bad-doc-types* (bad-doc-types))

(defparameter *doc-arglists-hash*
  (let ((hash (make-hash-table :test 'eq)))
    (dolist (spec *doc-file-arglists*)
      (setf (gethash (car spec) hash) spec))
    hash))

(defun doc-file-arglist (sym)
  (let ((spec (gethash sym *doc-arglists-hash*)))
    (when (and spec (listp (second spec)))
      (values (second spec) t))))

(defparameter *bad-doc-arglists* (check-arglists 'doc-file-arglist))

(defun print-bad-doc-arglists (stream)
  (dolist (a *bad-doc-arglists*)
    (let* ((sym (car a))
           (doc-file-arglist (second a))
           (real-arglist (third a))
           (file (fourth (gethash sym *doc-arglists-hash*))))
      (format stream "~&~s - ~a~%  Docs: ~s~%  Real: ~s"
              sym file doc-file-arglist real-arglist)
      (force-output stream))))

|#