/*
  update-subprims application-or-lib [subprims-file]
  copy the code section of subprims file onto the start of 
  the code section application-or-lib, if it fits.
  subprims-file defaults to "subprims", in the current directory.
  This is supposed to work for small patches; significant changes
  should require a full rebuild anyhow.
*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <CodeFragments.h>
#include "pef.h"
#include "lisp.h"


/*
  Open the indicated filename in the specified mode, reading and validating
  the file's PEF header.  Return the FILE pointer on success, NULL on failure.
*/

FILE *
open_PEF_file(char *name, char *mode, PEF_section *codeheader)
{
  PEF_header header;
  FILE *f = fopen(name, mode);
  if (f != NULL) {
    if ((fread(&header, sizeof(PEF_header), 1, f) != 1) ||
        (header.magic != sufficiently_advanced) ||
        (header.containerID != 'peff') ||
        (header.arch != kPowerPCCFragArch) ||
        (header.version != 1)) {
      fclose(f);
      f = NULL;
    } else {
      if ((fread(codeheader, sizeof(PEF_section), 1, f) != 1) ||
          (codeheader->region_kind != section_kind_code)) {
        fclose(f);
        f = NULL;
      }
    }
  }
  return f;
}

/*
  Read the raw contents of the PEF section from the file
  into a buffer and return that buffer.  Return NULL on failure.
  */
char *
read_section_contents(FILE *f, PEF_section *section)
{
  unsigned rawsize = section->rawsize;
  char *buffer = malloc(rawsize);
  
  if (buffer) {
    if (fseek(f, section->offset, SEEK_SET)) {
      free(buffer);
      buffer = NULL;
    } else {
      if (fread(buffer, rawsize, 1, f) != 1) {
        free(buffer);
        buffer = NULL;
      }
    }
  }
  return buffer;
}

int
patch_subprims(FILE *f, char *code, unsigned code_size_in_bytes, PEF_section *target_section)
{
  unsigned code_size_in_words = code_size_in_bytes>>2, target_elements, *words = (unsigned *) code;
  int status, target_subtag;
  LispObj header, nop;
  short desc = fileno(f);

  while (code_size_in_words && (words[code_size_in_words-1] == 0x60000000)) {
    code_size_in_words--;
    code_size_in_bytes-=4;
  }

  status = fseek(f, target_section->offset, SEEK_SET);
  if (status) {
    return status;
  }

  status = (fread(&header, sizeof(LispObj), 1, f) != 1);
  if (status) {
    return status;
  }

  status = (fread(&nop, sizeof(LispObj), 1, f) != 1);
  if (status) {
    return status;
  }

  if (nop != 0x60000000) {
    return 1;
  }

  target_subtag = header_subtag(header);
  if (target_subtag != subtag_code_vector) {
    return 1;
  }
  
  target_elements = header_element_count(header) - 1; /* account for leading NOP */
  if (target_elements < code_size_in_words) {
    /* Won't fit. */
    return 1;
  }

  lseek(desc, (target_section->offset)+8, SEEK_SET);
  return (write(desc, code, code_size_in_bytes) != code_size_in_bytes);

}


main(int argc, char **argv)
{
  FILE *subpfile, *targetfile;
  PEF_section subpcode, targetcode;
  char *subpname = ":subprims", *subpdata;

  if ((argc < 2) || (argc > 3)) {
    fprintf(stderr, "Usage: %s application-or-library [subprims]\n", argv[0]);
    exit(-1);
  }

  targetfile = open_PEF_file(argv[1], "r+", &targetcode);
  if (targetfile == NULL) {
    fprintf(stderr, "can't open PEF file %s for update.\n", argv[1]);
    exit(-1);
  }

  if (argc > 2) {
    subpname = argv[2];
  }

  subpfile = open_PEF_file(subpname, "r", &subpcode);
  if (subpfile == NULL) {
    fprintf(stderr, "can't open PEF file %s for reading.\n", subpname);
    fclose(targetfile);
    exit(-1);
  }

  subpdata = read_section_contents(subpfile, &subpcode);
  if (subpdata == NULL) {
    fprintf(stderr, "can't read section contents from file %s.\n", subpname);
    fclose(targetfile);
    fclose(subpfile);
    exit(-1);
  }

  if (patch_subprims(targetfile, subpdata, subpcode.rawsize, &targetcode)) {
    fprintf(stderr, "can't patch code in target file %s\n", argv[1]);
    fclose(targetfile);
    fclose(subpfile);
    exit(-1);
  }
  fclose(targetfile);
  fclose(subpfile);
  exit(0);
}

